/* From David Banks (1993), A Gallery of Surfaces in Four Dimensions 
   Knotted sphere, described as a surface of revolution generated by
   a knot in 3-space. The knot is the s-parameter curve in Xyz-space.

   X = sin(s/2) - 1.7*cos(2.0*s)*sin(s/2.0)*sin(s/2.0)*sin(s/2.0);
   x = cos(t) * X;
   y = -cos(s/2.0) + 2.0*sin(s)*sin(s/2)*sin(s/2);
   z = 1.5*cos(2.5*s)*sin(s/2)*sin(s/2);
   w = -sin(t)*X;

   Hui Ma, July 12, 1994
*/

#include <stdio.h>
#include <math.h>

main(int argc, char *argv[]) {
  FILE *out;
  int i, j, nu, nv;
  double du, dv, u, v;
  double X, x, y, z, w;
  double const_r, const_i;

  if(argc != 3) {
    printf("Usage : ks <nu> <nv>\n");
    exit(1);
  }
  nu = atoi(argv[1]);
  nv = atoi(argv[2]);

  printf("nu = %d, nv = %d\n", nu, nv);

  if ((out = fopen("ks.4mesh", "w")) == NULL) {
    fprintf(stderr, "error: Could not open file.\n");
    exit(1);
  }
  
  fprintf(out, "4vMESH\n%d %d\n",nu, nv);
  du = 2.0*M_PI / (double) (nu-1.0);
  dv = 2.0*M_PI / (double) (nv);

  v = 0.0;
  for (j=0; j<nv; j++) {
    u = 0.0;
    for (i=0; i<nu; i++) { 
      X = sin(u/2) - 1.7*cos(2.0*u)*sin(u/2.0)*sin(u/2.0)*sin(u/2.0);
      x = cos(v) * X;
      y = -cos(u/2.0) + 2.0*sin(u)*sin(u/2)*sin(u/2);
      z = 1.5*cos(2.5*u)*sin(u/2)*sin(u/2);
      w = -sin(v)*X;
/*      fprintf(out, "%f %f %f %f\n", x, y, z, w); */
      fprintf(out, "%f %f %f %f\n", x, y, w, z); 
      u = u + du;
    }
    v = v + dv;
  }
  fclose(out);
}
